# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Tests associated with the numerical propagators.

"""
import numpy as np
import pytest
from astropy import units as u
from astropy.coordinates import (
    GCRS,
    ITRS,
    CartesianDifferential,
    CartesianRepresentation,
    SkyCoord,
)
from astropy.time import Time

from satmad.propagation.numerical_propagators import ODESolverType
from satmad.propagation.tests.num_prop_analysis.num_prop_analysis_engine import (
    energy_along_trajectory,
    propagation_engine,
)


@pytest.fixture
def rv_init_leo_gcrs():
    """Initialises an Earth LEO satellite."""
    time = Time("2004-04-06T07:51:28.386009", scale="utc")

    # Vallado IAU 2000 - Table 3-6
    v_gcrs = CartesianDifferential(
        [-4.7432201610, 0.7905364950, 5.5337557240], unit=u.km / u.s
    )
    r_gcrs = CartesianRepresentation(
        [5102.50895290, 6123.01139910, 6378.13693380], unit=u.km
    )
    rv_init = SkyCoord(
        r_gcrs.with_differentials(v_gcrs),
        obstime=time,
        frame=GCRS,
        representation_type="cartesian",
        differential_type="cartesian",
    )

    return rv_init


def test_num_propagator(rv_init_leo_gcrs):
    """Tests the basic numerical propagation accuracy in terms of energy."""

    # Set up propagation config
    # Start with ITRS to test coord transform on propagator entry
    rv_init = rv_init_leo_gcrs.transform_to(ITRS)
    stepsize = 60 * u.s
    solver_type = ODESolverType.DOP853
    rtol = 1e-13
    atol = 1e-15
    init_time_offset = 0.5 * u.day
    duration = 10.001 * u.day

    # run propagation and get trajectory
    trajectory = propagation_engine(
        rv_init, stepsize, solver_type, init_time_offset, duration, rtol, atol
    )
    # print(trajectory.coord_list[-1])

    # Compute the energy along the trajectory
    time_list, energy_list, energy_diff_list = energy_along_trajectory(
        trajectory.coord_list
    )

    abs_energy_diff_list = abs(np.array(energy_diff_list))

    # Max energy diff (abs value) - should be low
    energy_diff_max = abs_energy_diff_list.max()
    # Mean energy diff - should be zero to have zero mean energy
    energy_diff_mean = np.array(energy_diff_list).mean()

    # initial energy diff and final energy diff to assess energy leak
    # First and last 100 elements taken and they should also ideally be zero
    init_energy_diff_mean = np.array(energy_diff_list)[:100].mean()
    final_energy_diff_mean = np.array(energy_diff_list)[-100:].mean()

    # print(f"max energy diff: {energy_diff_max}")
    # print(f"mean energy    : {energy_diff_mean}")
    #
    # print(f"init mean energy diff : {init_energy_diff_mean}")
    # print(f"final mean energy diff: {final_energy_diff_mean}")

    assert energy_diff_max < 4.50e-11
    assert energy_diff_mean < 2.95e-12

    assert init_energy_diff_mean < 1.25e-12
    assert final_energy_diff_mean < 5.20e-12



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Analysis of numerical propagation errors with DOP853 numerical propagation method.
Case: Position accuracy
"""
from typing import Dict

from astropy import units as u

from satmad.propagation.numerical_propagators import ODESolverType
from satmad.propagation.tests.num_prop_analysis.num_prop_analysis_engine import (
    analyse_pos_along_trajectory,
    init_rv,
    plot_pos_diff,
    plot_runtimes,
    propagation_engine,
    tle_leo,
)

if __name__ == "__main__":
    # Set up propagation config
    tle = tle_leo()
    rv_init = init_rv(tle, tle.epoch)
    stepsize = 120 * u.s
    solver_type = ODESolverType.DOP853
    rtol = 3e-14
    # atols = [1e-15, 1e-14, 1e-13, 1e-12, 1e-11, 1e-10]
    rtols = [1e-14, 1e-13, 1e-12, 1e-11, 1e-10]
    atol = 1e-14
    init_time_offset = 0 * u.day
    duration = 10.0 * u.day

    # atol effect for fixed low rtol: no impact on pos error
    # atol effect for fixed high rtol: no impact on pos error, lower accuracy
    # rtol effect for fixed low atol: pos error varies with rtol
    # rtol effect for fixed high atol: pos error varies with rtol, same accuracy

    # run propagation and get truth trajectory
    truth_traj = propagation_engine(
        rv_init,
        stepsize,
        ODESolverType.DOP853,
        init_time_offset,
        duration,
        rtol=3e-14,
        atol=1e-15,
    )

    cases: Dict[str, Dict] = {}
    type(cases)

    # run all cases
    for rtol in rtols:
        name = str(rtol)
        # rtol = atol

        analyse_pos_along_trajectory(
            truth_traj,
            cases,
            name,
            rv_init,
            stepsize,
            solver_type,
            init_time_offset,
            duration,
            rtol,
            atol,
        )

    session_name = f"[DOP853, atol = {atol} , varying rtol]"

    # Plot pos diff differences
    plot_pos_diff(cases, title=session_name, logscale_plot=True)

    # Plot runtime for one case to see its behaviour
    plot_pos_diff(
        {"1e-10": cases["1e-10"]}, title=f"{session_name}", logscale_plot=False
    )

    # Plot runtimes
    plot_runtimes(
        cases,
        title=f"{session_name} [duration: {duration.to_value(u.day)} days]",
        logscale_plot=True,
    )



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Analysis of numerical propagation errors with DOP853 numerical propagation method.
Case: Eccentricity and pos accuracy.
"""
from typing import Dict

from astropy import units as u

from satmad.propagation.numerical_propagators import ODESolverType
from satmad.propagation.tests.num_prop_analysis.num_prop_analysis_engine import (
    analyse_pos_along_trajectory,
    init_rv,
    plot_pos_diff,
    plot_runtimes,
    propagation_engine,
    tle_heo,
)

if __name__ == "__main__":

    # Set up propagation config
    stepsize = 120 * u.s
    solver_type = ODESolverType.DOP853
    rtol = 1e-11
    atol = 1e-11
    init_time_offset = 0 * u.day
    duration = 10.0 * u.day

    ecc_list = [0.001, 0.1, 0.2, 0.4, 0.5, 0.7, 0.75]
    tle_list = []
    for ecc in ecc_list:
        tle = tle_heo()
        tle.eccentricity = ecc
        tle_list.append(tle)

    cases: Dict[str, Dict] = {}
    type(cases)

    # run all cases
    for tle in tle_list:
        name = str(tle.eccentricity)

        rv_init = init_rv(tle, tle.epoch)

        # run propagation and get truth trajectory
        truth_traj = propagation_engine(
            rv_init,
            stepsize,
            ODESolverType.DOP853,
            init_time_offset,
            duration,
            rtol=3e-14,
            atol=1e-15,
        )

        analyse_pos_along_trajectory(
            truth_traj,
            cases,
            name,
            rv_init,
            stepsize,
            solver_type,
            init_time_offset,
            duration,
            rtol,
            atol,
        )

    session_name = f"[DOP853, atol = {atol} , rtol = {rtol}]"

    # Plot pos differences
    plot_pos_diff(cases, title=session_name, logscale_plot=True)

    # Plot runtimes
    plot_runtimes(
        cases,
        title=f"{session_name} [duration: {duration.to_value(u.day)} days]",
        logscale_plot=False,
    )



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Ground and Earth Location classes. These are very similar to
:class:`astropy.coordinates.EarthLocation` class.

"""

# class GroundLocation():
#
# class EarthLocation(GroundLocation):



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Analysis of numerical propagation errors with various numerical propagation methods.
Case: Energy error and runtimes.
"""
from typing import Dict

from astropy import units as u

from satmad.propagation.numerical_propagators import ODESolverType
from satmad.propagation.tests.num_prop_analysis.num_prop_analysis_engine import (
    analyse_energy_along_trajectory,
    init_rv,
    plot_energy_diff,
    plot_runtimes,
    tle_heo,
)

if __name__ == "__main__":
    # Set up propagation config
    # tle = tle_leo()
    tle = tle_heo()
    rv_init = init_rv(tle, tle.epoch)
    stepsize = 120 * u.s
    solver_types = [solver_type for solver_type in ODESolverType]
    # solver_types.remove(ODESolverType.BDF)
    # solver_types.remove(ODESolverType.RK23)
    # solver_types.clear()
    # solver_types.append(ODESolverType.DOP853)
    rtol = 3e-14
    # atol = [1e-12, 1e-12, 1e-12, 1e-15, 1e-15, 1e-15]
    atol = 1e-12
    init_time_offset = 0 * u.day
    duration = 1.0 * u.day

    cases: Dict[str, Dict] = {}
    type(cases)

    # run all cases
    for solver_type in solver_types:
        name = solver_type.value

        analyse_energy_along_trajectory(
            cases,
            name,
            rv_init,
            stepsize,
            solver_type,
            init_time_offset,
            duration,
            rtol,
            atol,
        )

    session_name = f"[rtol: {rtol}, atol: {atol}]"

    # Plot energy differences
    plot_energy_diff(cases, title=session_name, logscale_plot=True)

    # Plot DOP853 on linear scale to see its behaviour
    plot_energy_diff(
        {"DOP853": cases["DOP853"]}, title=session_name, logscale_plot=False
    )

    # Plot runtimes
    plot_runtimes(
        cases, title=f"{session_name} [duration: {duration.to_value(u.day)} days]"
    )
